//   Copyright 2020 Esri
//   Licensed under the Apache License, Version 2.0 (the "License");
//   you may not use this file except in compliance with the License.
//   You may obtain a copy of the License at
//
//   https://www.apache.org/licenses/LICENSE-2.0
//
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS,
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and
//   limitations under the License.

#include "Display_a_map.h"

#include "ArcGISRuntimeEnvironment.h"
#include "Map.h"
#include "MapQuickView.h"

#include "Point.h"
#include "RouteTask.h"
#include "RouteParameters.h"
#include "SimpleLineSymbol.h"
#include "SimpleMarkerSymbol.h"
#include "Stop.h"

#include <QUrl>

using namespace Esri::ArcGISRuntime;

Display_a_map::Display_a_map(QObject* parent /* = nullptr */):
  QObject(parent),

  m_map(new Map(BasemapStyle::ArcGISStreets, this)),

  m_currentState(RouteBuilderStatus::NotStarted)

{

  setupRouteTask();

}

Display_a_map::~Display_a_map()
{
}

MapQuickView* Display_a_map::mapView() const
{
  return m_mapView;
}

void Display_a_map::setupViewpoint()
{

  const Point center(-118.24532, 34.05398, SpatialReference::wgs84());
  const Viewpoint viewpoint(center, 144447.638572);

  m_mapView->setViewpoint(viewpoint);

  connect(m_mapView, &MapQuickView::mouseClicked, this, [this](QMouseEvent mouse)
  {
    const Point mapPoint = m_mapView->screenToLocation(mouse.x(), mouse.y());

    switch (m_currentState)
    {
      case RouteBuilderStatus::NotStarted:
        resetState();
        m_currentState = RouteBuilderStatus::SelectedStart;
        m_startGraphic->setGeometry(mapPoint);
        break;
      case RouteBuilderStatus::SelectedStart:
        m_currentState = RouteBuilderStatus::SelectedStartAndEnd;
        m_endGraphic->setGeometry(mapPoint);
        findRoute();
        break;

      case RouteBuilderStatus::SelectedStartAndEnd:
        // Ignore touches while routing is in progress
        break;
    }
  });

  m_graphicsOverlay = new GraphicsOverlay(this);
  m_mapView->graphicsOverlays()->append(m_graphicsOverlay);

  SimpleLineSymbol* startOutlineSymbol = new SimpleLineSymbol(SimpleLineSymbolStyle::Solid, QColor("blue"), 2/*width*/, this);
  SimpleMarkerSymbol* startSymbol = new SimpleMarkerSymbol(SimpleMarkerSymbolStyle::Diamond, QColor("orange"), 12/*width*/, this);
  startSymbol->setOutline(startOutlineSymbol);
  m_startGraphic = new Graphic(this);
  m_startGraphic->setSymbol(startSymbol);

  SimpleLineSymbol* endOutlineSymbol = new SimpleLineSymbol(SimpleLineSymbolStyle::Solid, QColor("red"), 2/*width*/, this);
  SimpleMarkerSymbol* endSymbol = new SimpleMarkerSymbol(SimpleMarkerSymbolStyle::Square, QColor("green"), 12/*width*/, this);
  endSymbol->setOutline(endOutlineSymbol);
  m_endGraphic = new Graphic(this);
  m_endGraphic->setSymbol(endSymbol);

  SimpleLineSymbol* lineSymbol = new SimpleLineSymbol(SimpleLineSymbolStyle::Solid, QColor("blue"), 4/*width*/, this);
  m_lineGraphic = new Graphic(this);
  m_lineGraphic->setSymbol(lineSymbol);

  m_graphicsOverlay->graphics()->append(QList<Graphic*> {m_startGraphic, m_endGraphic, m_lineGraphic});

}

void Display_a_map::setupRouteTask()
{
  // create the route task pointing to an online service
  m_routeTask = new RouteTask(QUrl("https://route-api.arcgis.com/arcgis/rest/services/World/Route/NAServer/Route_World"), this);

  // connect to createDefaultParametersCompleted signal
  connect(m_routeTask, &RouteTask::createDefaultParametersCompleted, this, [this](QUuid, RouteParameters routeParameters)
  {
    // Store the resulting route parameters
    m_routeParameters = routeParameters;
  });

  // connect to solveRouteCompleted signal
  connect(m_routeTask, &RouteTask::solveRouteCompleted, this, [this](QUuid, RouteResult routeResult)
  {
    // Add the route graphic once the solve completes
    const Route generatedRoute = routeResult.routes().at(0);
    m_lineGraphic->setGeometry(generatedRoute.routeGeometry());
    m_currentState = RouteBuilderStatus::NotStarted;

    // set the direction maneuver list model
    m_directions = generatedRoute.directionManeuvers(this);
    emit directionsChanged();
  });

  // create the default parameters which will load the route task implicitly
  m_routeTask->createDefaultParameters();
}

void Display_a_map::findRoute()
{
  if (m_routeTask->loadStatus() != LoadStatus::Loaded || m_routeParameters.isEmpty())
    return;

  // set parameters to return directions
  m_routeParameters.setReturnDirections(true);

  // clear previous stops from the parameters
  m_routeParameters.clearStops();

  // set the stops to the parameters
  const Stop stop1(m_startGraphic->geometry());
  const Stop stop2(m_endGraphic->geometry());
  m_routeParameters.setStops(QList<Stop> { stop1, stop2 });

  // solve the route with the parameters
  m_routeTask->solveRoute(m_routeParameters);
}

void Display_a_map::resetState()
{
  m_startGraphic->setGeometry(Point());
  m_endGraphic->setGeometry(Point());
  m_lineGraphic->setGeometry(Point());
  m_directions = nullptr;
  m_currentState = RouteBuilderStatus::NotStarted;
}

// Set the view (created in QML)
void Display_a_map::setMapView(MapQuickView* mapView)
{
  if (!mapView || mapView == m_mapView)
  {
    return;
  }
  m_mapView = mapView;
  m_mapView->setMap(m_map);
  setupViewpoint();
  emit mapViewChanged();
}

